/*
 * Copyright (c) 2007 Alexey Vatchenko <av@bsdua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cdma.h"
#include "util.h"
#include "cmotech_ccu550.h"

/*
 * This modem has one "feature" that requires special handling. It's SMS Alert.
 * $$SMSALERT\r\n response can be supplied by modem as an answer to any AT
 * command after response to the exact AT command. E.g. :
 *	AT$$TIME
 *	$$TIME: 20080102,173026,WED
 *	OK
 *	$$SMSALERT
 *	OK
 * Actually, not only $$SMSALERT... Other commands have this ability too.
 * A lot of them :)
 */

int
cmotech_ccu550_getlimit(struct cdma *ctx, enum cdma_limit what)
{
	switch (what) {
	case CDMA_LIMIT_PHONE:
		return (20);
	case CDMA_LIMIT_CONTACT:
		return (0);
	}
	return (0);
}

int
cmotech_ccu550_gettime(struct cdma *ctx, struct cdma_time *ct)
{
	size_t sz, i;
	char *p, *nxt;
	char mask[] = {
		'd', 'd', 'd', 'd', 'd', 'd', 'd', 'd', ',',
		'd', 'd', 'd', 'd', 'd', 'd', ',',
		'c', 'c', 'c'
	};
	int status, error;

	error = send_cmd(ctx->fd, "AT$$TIME\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_tagged_line(ctx->buf, sz, "$$TIME: ", &nxt);
	if (p == NULL) {
		warnx("can't determine time");
		return (-1);
	}

	/* YYYY-MM-DD,HH:MM:SS,WKD */
	p += 8;			/* tag length */
	if ((size_t)(nxt - p) < sizeof(mask)) {
		warnx("invalid device response");
		return (-1);
	}

	/* validate format */
	error = 0;
	for (i = 0; i < sizeof(mask); i++) {
		switch (mask[i]) {
		case 'd':	/* digit */
			if (!isdigit(p[i]))
				error = -1;
			break;
		case 'c':	/* digit */
			if (!isupper(p[i]))
				error = -1;
			break;
		}
		if (error != 0) {
			warnx("invalid time format");
			return (-1);
		}
	}

	ct->year = p[3] - '0' + 10 * (p[2] - '0') + 100 * (p[1] - '0') +
	    1000 * (p[0] - '0');
	ct->month = 10 * (p[4] - '0') + (p[5] - '0');
	if (ct->month == 0)
		ct->month = 1;	/* probably starts with 0 */
	ct->day = 10 * (p[6] - '0') + (p[7] - '0');
	ct->hour = 10 * (p[9] - '0') + (p[10] - '0');
	ct->min = 10 * (p[11] - '0') + (p[12] - '0');
	ct->sec = 10 * (p[13] - '0') + (p[14] - '0');
	memcpy(ct->wday, p + 16, 3);
	ct->wday[0] = p[16];
	ct->wday[1] = tolower(p[17]);
	ct->wday[2] = tolower(p[18]);
	ct->wday[3] = '\0';
	return (0);
}

int
cmotech_ccu550_pininfo(struct cdma *ctx, int *flags)
{
	char *p, *s, *nxt;
	size_t sz;
	int status, error;

	error = send_cmd(ctx->fd, "AT$$CHV1?\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_tagged_line(ctx->buf, sz, "$$CHV1: ", &nxt);
	if (p == NULL) {
		warnx("can't determine PIN code status");
		return (-1);
	}

	s = get_answer_array(p, p - ctx->buf + sz, 4, "$$CHV1");
	if (s == NULL) {
		warnx("can't determine PIN code status");
		return (-1);
	}

	*flags = 0;
	p = s + strlen(s) + 1;
	if (strcmp(p, "1") == 0)
		*flags |= PIN_PUK;
	p = p + strlen(p) + 1;
	if (strcmp(p, "1") == 0)
		*flags |= PIN_ENABLED;
	p = p + strlen(p) + 1;
	if (strcmp(p, "1") == 0)
		*flags |= PIN_NEEDED;

	return (0);
}

int
cmotech_ccu550_pin_enable(struct cdma *ctx, const char *pin, int toenable)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$%s=%s\r\n",
	    (toenable) ? "ENACHV1" : "DISCHV1", pin);
	if (error == -1) {
		warnx("can't enable/disable PIN code");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't enable/disable PIN code");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$%s=%s\r\n",
	    (toenable) ? "ENACHV1" : "DISCHV1", pin);
	if (error == -1) {
		warnx("can't enable/disable PIN code");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu550_pin_change(struct cdma *ctx, const char *old, const char *pin)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$CHGCHV1=%s,%s\r\n", old, pin);
	if (error == -1) {
		warnx("can't change PIN code");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't change PIN code");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$CHGCHV1=%s,%s\r\n", old, pin);
	if (error == -1) {
		warnx("can't change PIN code");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu550_pin_enter(struct cdma *ctx, const char *pin)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$CHV1=%s\r\n", pin);
	if (error == -1) {
		warnx("can't enter PIN code");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't enter PIN code");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$CHV1=%s\r\n", pin);
	if (error == -1) {
		warnx("can't enter PIN code");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu550_puk_enter(struct cdma *ctx, const char *puk, const char *pin)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$PUKCHV1=%s,%s\r\n", puk, pin);
	if (error == -1) {
		warnx("can't enter PUK code");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't enter PUK code");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$PUKCHV1=%s,%s\r\n", puk, pin);
	if (error == -1) {
		warnx("can't enter PUK code");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu550_getmode(struct cdma *ctx)
{
	char *p, *s, *nxt;
	size_t sz;
	int status, error;

	error = send_cmd(ctx->fd, "AT$$HDRMD?\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_tagged_line(ctx->buf, sz, "$$HDRMD: ", &nxt);
	if (p == NULL) {
		warnx("can't determine mode");
		return (-1);
	}

	s = get_answer_array(p, p - ctx->buf + sz, 1, "$$HDRMD");
	if (s == NULL) {
		warnx("can't determine mode");
		return (-1);
	}

	if (strcmp(s, "0") == 0) {
		printf("Hybrid (1xEV-DO)\n");
		return (0);
	} else if (strcmp(s, "1") == 0) {
		printf("1x\n");
		return (1);
	} else if (strcmp(s, "2") == 0) {
		printf("EV-DO\n");
		return (2);
	}

	printf("Unknown\n");
	return (-1);
}

int
cmotech_ccu550_setmode(struct cdma *ctx, int mode)
{
	char cmd[20];
	size_t sz;
	int status, error;

	error = snprintf(cmd, sizeof(cmd), "AT$$HDRMD=%1u,1\r\n", mode);
	if (error == -1 || (size_t)error >= sizeof(cmd)) {
		warnx("can't set mode");
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu550_sms_send(struct cdma *ctx, const char *dst, const char *msg,
    size_t msgsz)
{
	char *cmd;
	char *p, *nxt, *s;
	size_t sz, cmdsz, i;
	int status, error;

	/* Check for unicode */
	for (i = 0; i < msgsz; i++) {
		if ((u_char)msg[i] > 0x7f) {
			warnx("UNICODE is not supported yet");
			return (-1);
		}
	}

	cmdsz = 11 + strlen(dst) + 4 + msgsz + 2;
	cmd = (char *)malloc(cmdsz + 1);
	if (cmd == NULL) {
		warn("malloc");
		return (-1);
	}

	error = snprintf(cmd, cmdsz, "AT$$MOSEND=%s,,2,", dst);
	if (error == -1 || (size_t)error >= cmdsz) {
		warnx("can't send SMS");
		free(cmd);
		return (-1);
	}

	p = (u_char *)cmd + 11 + strlen(dst) + 4;
	memcpy(p, msg, msgsz);
	p += msgsz;
	*p++ = '\r';
	*p++ = '\n';
	*p = '\0';

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	/*
	 * Read MOSEND status. If it's equal to 1 read MOACK status.
	 */
	p = get_tagged_line(ctx->buf, sz, "$$MOSEND: ", &nxt);
	if (p == NULL) {
		warnx("invalid response from device");
		return (-1);
	}

	s = get_answer_array(p, sz - (p - ctx->buf), 1, "$$MOSEND");
	if (s == NULL) {
		warnx("invalid response from device");
		return (-1);
	}

	if (*s != '1' || *(s + 1) != '\0') {
		warnx("can't send SMS");
		return (-1);
	}

	/* Read $$MOACK */
	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_tagged_line(ctx->buf, sz, "$$MOACK: ", &nxt);
	if (p == NULL) {
		warnx("invalid response from device");
		return (-1);
	}

	s = get_answer_array(p, sz - (p - ctx->buf), 1, "$$MOACK");
	if (s == NULL) {
		warnx("invalid response from device");
		return (-1);
	}

	if (*s != '1' || *(s + 1) != '\0') {
		warnx("can't send SMS");
		return (-1);
	}

	return (0);
}

int
cmotech_ccu550_sms_read(struct cdma *ctx, struct cdma_sms *sms)
{
	char *p, *s, *nxt;
	size_t sz;
	int status, error;

	error = send_cmd(ctx->fd, "AT$$READMT=0\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_tagged_line(ctx->buf, sz, "$$READMT: ", &nxt);
	if (p == NULL) {
		warnx("there is no unread SMS");
		return (0);
	}

	s = get_answer_array(p, p - ctx->buf + sz, 6, "$$READMT");
	if (s == NULL) {
		warnx("can't read SMS");
		return (-1);
	}

	memset(sms, 0, sizeof(*sms));

	s += strlen(s) + 1;
	/* suppose that format is CCYYMMDDhhmmss */
	if (strlen(s) == 14) {
		sms->when.year = (s[0] - '0') * 1000 + (s[1] - '0') * 100 +
		    (s[2] - '0') * 10 + (s[3] - '0');
		sms->when.month = (s[4] - '0') * 10 + (s[5] - '0');
		if (sms->when.month == 0)
			sms->when.month = 1;	/* probably starts with 0 */
		sms->when.day = (s[6] - '0') * 10 + (s[7] - '0');
		sms->when.hour = (s[8] - '0') * 10 + (s[9] - '0');
		sms->when.min = (s[10] - '0') * 10 + (s[11] - '0');
		sms->when.sec = (s[12] - '0') * 10 + (s[13] - '0');
	}

	s += strlen(s) + 1;
	strlcpy(sms->from, s, sizeof(sms->from));

	s += strlen(s) + 1;
	s += strlen(s) + 1;
	if (strcmp(s, "2") != 0) {
		warnx("UNICODE is not supported yet");
		return (-1);
	}

	s += strlen(s) + 1;
	strlcpy(sms->msg, s, sizeof(sms->msg));
	return (1);
}

int
cmotech_ccu550_sms_check(struct cdma *ctx, u_int *n_sms)
{
	warnx("Not implemented");
	return (-1);
}

int
cmotech_ccu550_sms_enum(struct cdma *ctx, cdma_sms_callback cb)
{
	char *p, *s, *nxt;
	size_t sz, sz2;
	int mb, status, error;
	struct cdma_sms sms;
	char *idx;

	error = send_cmd(ctx->fd, "AT$$RDAUSMS\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = ctx->buf;
	sz2 = sz;
	while (p != NULL) {
		p = get_tagged_line(p, sz2, "$$RDAUSMS: ", &nxt);
		if (p == NULL)
			break;
		sz2 = sz - (p - ctx->buf);

		s = get_answer_array(p, sz2, 8, "$$RDAUSMS");
		if (s == NULL) {
			warnx("can't read stored SMS");
			return (-1);
		}

		memset(&sms, 0, sizeof(sms));

		idx = s;
		s = s + strlen(s) + 1;
		s = s + strlen(s) + 1;
		s = s + strlen(s) + 1;
		if (strlen(s) == 14) {
			sms.when.year = (s[0] - '0') * 1000 +
			    (s[1] - '0') * 100 + (s[2] - '0') * 10 +
			    (s[3] - '0');
			sms.when.month = (s[4] - '0') * 10 + (s[5] - '0');
			if (sms.when.month == 0)
				sms.when.month = 1;
			sms.when.day = (s[6] - '0') * 10 + (s[7] - '0');
			sms.when.hour = (s[8] - '0') * 10 + (s[9] - '0');
			sms.when.min = (s[10] - '0') * 10 + (s[11] - '0');
			sms.when.sec = (s[12] - '0') * 10 + (s[13] - '0');
		}
		s = s + strlen(s) + 1;
		strlcpy(sms.from, s, sizeof(sms.from));
		s = s + strlen(s) + 1;

		s = s + strlen(s) + 1;
		mb = 1;
		if (strcmp(s, "2") == 0)
			mb = 0;

		s = s + strlen(s) + 1;
		if (!mb) {
			strlcpy(sms.msg, s, sizeof(sms.msg));
			cb(idx, &sms);
		} else
			warnx("(unicode is not supported yet)\n");

		p = nxt;
	}
	return (0);
}

int
cmotech_ccu550_sms_delete(struct cdma *ctx, u_int index)
{
	char cmd[20];
	size_t sz;
	int status, error;

	error = snprintf(cmd, sizeof(cmd), "AT$$DELUSMS=%u\r\n", index);
	if (error == -1 || (size_t)error >= sizeof(cmd)) {
		warnx("can't delete stored SMS");
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	return (0);
}

int
cmotech_ccu550_sms_save(struct cdma *ctx, const char *phone,
    struct cdma_time *ct, const char *msg)
{
	char *cmd;
	size_t sz, msgsz, i;
	int status, error;

	msgsz = strlen(msg);

	/* Check for unicode */
	for (i = 0; i < msgsz; i++) {
		if ((u_char)msg[i] > 0x7f) {
			warnx("UNICODE is not supported yet");
			return (-1);
		}
	}

	error = snprintf(NULL, 0,
	    "AT$$PCSMS2UIM=1,%04d%02d%02d%02d%02d%02d,%s,%s\r\n", ct->year,
	    ct->month, ct->day, ct->hour, ct->min, ct->sec, phone, msg);
	if (error == -1) {
		warnx("can't save SMS");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("malloc");
		return (-1);
	}

	error = snprintf(cmd, error + 1,
	    "AT$$PCSMS2UIM=1,%04d%02d%02d%02d%02d%02d,%s,%s\r\n", ct->year,
	    ct->month, ct->day, ct->hour, ct->min, ct->sec, phone, msg);
	if (error == -1) {
		warnx("can't save SMS");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	return (0);
}

int
cmotech_ccu550_pb_enum(struct cdma *ctx, cdma_pb_callback cb)
{
	char *p, *s, *nxt;
	size_t sz, sz2;
	int status, error;
	char *idx, *phone, *name;

	error = send_cmd(ctx->fd, "AT$$UPBRD=0\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = ctx->buf;
	sz2 = sz;
	while (p != NULL) {
		p = get_tagged_line(p, sz2, "$$UPBRD: ", &nxt);
		if (p == NULL)
			break;
		sz2 = sz - (p - ctx->buf);

		s = get_answer_array(p, sz2, 4, "$$UPBRD");
		if (s == NULL) {
			warnx("can't read phone book");
			return (-1);
		}

		idx = s;
		s = s + strlen(s) + 1;
		phone = s;
		s = s + strlen(s) + 1;
		s = s + strlen(s) + 1;
		name = s;

		cb(idx, phone, name);

		p = nxt;
	}

	return (0);
}

/* NOTE: this modem has limitation of 20 symbols per phone number */
int
cmotech_ccu550_pb_add(struct cdma *ctx, u_int index, const char *phone,
    const char *name)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$UPBWR=%u,%s,2,%s\r\n", index, phone,
	    name);
	if (error == -1) {
		warnx("can't add phonebook record");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't add phonebook record");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$UPBWR=%u,%s,2,%s\r\n", index,
	    phone, name);
	if (error == -1) {
		warnx("can't add phonebook record");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu550_pb_delete(struct cdma *ctx, u_int index)
{
	char cmd[20];
	size_t sz;
	int status, error;

	error = snprintf(cmd, sizeof(cmd), "AT$$PBDEL=%u\r\n", index);
	if (error == -1 || (size_t)error >= sizeof(cmd)) {
		warnx("can't delete phonebook record");
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}
