/*
 * Copyright (c) 2007 Alexey Vatchenko <av@bsdua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cdma.h"
#include "util.h"
#include "cmotech_ccu650u.h"
#include "cmotech_ccu550.h"

int
cmotech_ccu650u_getlimit(struct cdma *ctx, enum cdma_limit what)
{
	switch (what) {
	case CDMA_LIMIT_PHONE:
	case CDMA_LIMIT_CONTACT:
		return (20);
	}
	return (0);
}

int
cmotech_ccu650u_gettime(struct cdma *ctx, struct cdma_time *ct)
{
	size_t sz, i;
	char *p, *nxt;
	char mask[] = {
		'd', 'd', 'd', 'd', '-', 'd', 'd', '-', 'd', 'd', ',',
		'd', 'd', ':', 'd', 'd', ':', 'd', 'd', ',',
		'c', 'c', 'c'
	};
	int status, error;

	error = send_cmd(ctx->fd, "AT$$TIME\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_line(ctx->buf, sz, &nxt);
	if (p == NULL) {
		warnx("can't determine time");
		return (-1);
	}

	/* Skip prefix "$$TIME: " */
	if (nxt - p < 8) {
		warnx("invalid device response");
		return (-1);
	}

	/* YYYY-MM-DD,HH:MM:SS,WKD */
	p += 8;
	if ((size_t)(nxt - p) < sizeof(mask)) {
		warnx("invalid device response");
		return (-1);
	}

	/* validate format */
	error = 0;
	for (i = 0; i < sizeof(mask); i++) {
		switch (mask[i]) {
		case 'd':	/* digit */
			if (!isdigit(p[i]))
				error = -1;
			break;
		case 'c':	/* digit */
			if (!isupper(p[i]))
				error = -1;
			break;
		case '-':
			if (p[i] != '-')
				error = -1;
			break;
		case ':':
			if (p[i] != ':')
				error = -1;
			break;
		}
		if (error != 0) {
			warnx("invalid time format");
			return (-1);
		}
	}

	ct->year = p[3] - '0' + 10 * (p[2] - '0') + 100 * (p[1] - '0') +
	    1000 * (p[0] - '0');
	ct->month = 10 * (p[5] - '0') + (p[6] - '0');
	if (ct->month == 0)
		ct->month = 1;	/* probably starts with 0 */
	ct->day = 10 * (p[8] - '0') + (p[9] - '0');
	ct->hour = 10 * (p[11] - '0') + (p[12] - '0');
	ct->min = 10 * (p[14] - '0') + (p[15] - '0');
	ct->sec = 10 * (p[17] - '0') + (p[18] - '0');
	memcpy(ct->wday, p + 20, 3);
	ct->wday[0] = p[20];
	ct->wday[1] = tolower(p[21]);
	ct->wday[2] = tolower(p[22]);
	ct->wday[3] = '\0';
	return (0);
}

int
cmotech_ccu650u_pininfo(struct cdma *ctx, int *flags)
{
	return cmotech_ccu550_pininfo(ctx, flags);
}

int
cmotech_ccu650u_pin_enable(struct cdma *ctx, const char *pin, int toenable)
{
	return cmotech_ccu550_pin_enable(ctx, pin, toenable);
}

int
cmotech_ccu650u_pin_change(struct cdma *ctx, const char *old, const char *pin)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$CHGCHV=1,%s,%s\r\n", old, pin);
	if (error == -1) {
		warnx("can't change PIN code");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't change PIN code");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$CHGCHV=1,%s,%s\r\n", old, pin);
	if (error == -1) {
		warnx("can't change PIN code");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu650u_pin_enter(struct cdma *ctx, const char *pin)
{
	return cmotech_ccu550_pin_enter(ctx, pin);
}

int
cmotech_ccu650u_puk_enter(struct cdma *ctx, const char *puk, const char *pin)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$PUKCHV=1,%s,%s\r\n", puk, pin);
	if (error == -1) {
		warnx("can't enter PUK code");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't enter PUK code");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$PUKCHV=1,%s,%s\r\n", puk, pin);
	if (error == -1) {
		warnx("can't enter PUK code");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu650u_getmode(struct cdma *ctx)
{
	char *p, *s, *nxt;
	size_t sz;
	int status, error;

	error = send_cmd(ctx->fd, "AT$$HDR_MODE?\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_line(ctx->buf, sz, &nxt);
	if (p == NULL) {
		warnx("can't determine mode");
		return (-1);
	}

	s = get_answer_array(p, p - ctx->buf + sz, 1, "$$HDR_MODE");
	if (s == NULL) {
		warnx("can't determine mode");
		return (-1);
	}

	if (strcmp(s, "0") == 0) {
		printf("Hybrid (1xEV-DO)\n");
		return (0);
	} else if (strcmp(s, "1") == 0) {
		printf("1x\n");
		return (1);
	} else if (strcmp(s, "2") == 0) {
		printf("EV-DO\n");
		return (2);
	}

	printf("Unknown\n");
	return (-1);
}

int
cmotech_ccu650u_setmode(struct cdma *ctx, int mode)
{
	char cmd[20];
	size_t sz;
	int status, error;

	error = snprintf(cmd, sizeof(cmd), "AT$$HDR_MODE=%1u\r\n", mode);
	if (error == -1 || (size_t)error >= sizeof(cmd)) {
		warnx("can't set mode");
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

static const char hex_alpha[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8',
    '9', 'A', 'B', 'C', 'D', 'E', 'F'};

int
cmotech_ccu650u_sms_send(struct cdma *ctx, const char *dst, const char *msg,
    size_t msgsz)
{
	char *cmd;
	u_char *p;
	size_t sz, cmdsz, i;
	int status, error;

	/* Check for unicode */
	for (i = 0; i < msgsz; i++) {
		if ((u_char)msg[i] > 0x7f) {
			warnx("UNICODE is not supported yet");
			return (-1);
		}
	}

	cmdsz = 11 + 2 * strlen(dst) + 11 + msgsz * 2 + 2;
	cmd = (char *)malloc(cmdsz + 1);
	if (cmd == NULL) {
		warn("malloc");
		return (-1);
	}

	error = snprintf(cmd, cmdsz, "AT$$MOSEND=%s,%s,4098,0,2,", dst, dst);
	if (error == -1 || (size_t)error >= cmdsz) {
		warnx("can't send SMS");
		free(cmd);
		return (-1);
	}

	p = (u_char *)cmd + 11 + 2 * strlen(dst) + 11;
	for (i = 0; i < msgsz; i++) {
		*p++ = hex_alpha[(u_char)msg[i] / 16];
		*p++ = hex_alpha[(u_char)msg[i] % 16];
	}
	*p++ = '\r';
	*p++ = '\n';
	*p = '\0';

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	return (0);
}

int
cmotech_ccu650u_sms_read(struct cdma *ctx, struct cdma_sms *sms)
{
	return cmotech_ccu550_sms_read(ctx, sms);
}

int
cmotech_ccu650u_sms_check(struct cdma *ctx, u_int *n_sms)
{
	char *p, *s, *nxt;
	size_t sz;
	const char *errstr;
	int status, error;

	error = send_cmd(ctx->fd, "AT$$MTCNT\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = get_line(ctx->buf, sz, &nxt);
	if (p == NULL) {
		warnx("can't check for new SMS");
		return (-1);
	}

	s = get_answer_array(p, p - ctx->buf + sz, 2, "$$MTCNT");
	if (s == NULL) {
		warnx("can't check for new SMS");
		return (-1);
	}

	s = s + strlen(s) + 1;
	*n_sms = strtonum(s, 0, UINT_MAX, &errstr);
	if (errstr != NULL) {
		warnx("invalid response");
		return (-1);
	}

	return (0);
}

int
cmotech_ccu650u_sms_enum(struct cdma *ctx, cdma_sms_callback cb)
{
	char *p, *s, *nxt;
	size_t sz, sz2;
	int mb, status, error;
	struct cdma_sms sms;
	char *idx;

	error = send_cmd(ctx->fd, "AT$$UIMREADALL\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = ctx->buf;
	sz2 = sz;
	while (p != NULL) {
		p = get_line(p, sz2, &nxt);
		if (p == NULL)
			break;
		sz2 = sz - (p - ctx->buf);

		s = get_answer_array(p, sz2, 7, "$$UIMREADALL");
		if (s == NULL) {
			warnx("can't read stored SMS");
			return (-1);
		}

		memset(&sms, 0, sizeof(sms));

		idx = s;
		s = s + strlen(s) + 1;
		s = s + strlen(s) + 1;
		if (strlen(s) == 14) {
			sms.when.year = (s[0] - '0') * 1000 +
			    (s[1] - '0') * 100 + (s[2] - '0') * 10 +
			    (s[3] - '0');
			sms.when.month = (s[4] - '0') * 10 + (s[5] - '0');
			if (sms.when.month == 0)
				sms.when.month = 1;
			sms.when.day = (s[6] - '0') * 10 + (s[7] - '0');
			sms.when.hour = (s[8] - '0') * 10 + (s[9] - '0');
			sms.when.min = (s[10] - '0') * 10 + (s[11] - '0');
			sms.when.sec = (s[12] - '0') * 10 + (s[13] - '0');
		}
		s = s + strlen(s) + 1;
		strlcpy(sms.from, s, sizeof(sms.from));
		s = s + strlen(s) + 1;

		s = s + strlen(s) + 1;
		mb = 0;
		if (strcmp(s, "4") == 0)
			mb = 1;

		s = s + strlen(s) + 1;
		if (!mb) {
			strlcpy(sms.msg, s, sizeof(sms.msg));
			cb(idx, &sms);
		} else
			warnx("(unicode is not supported yet)\n");

		p = nxt;
	}
	return (0);
}

int
cmotech_ccu650u_sms_delete(struct cdma *ctx, u_int index)
{
	char cmd[20];
	size_t sz;
	int status, error;

	error = snprintf(cmd, sizeof(cmd), "AT$$UIMDEL=%u\r\n", index);
	if (error == -1 || (size_t)error >= sizeof(cmd)) {
		warnx("can't delete stored SMS");
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	return (0);
}

int
cmotech_ccu650u_sms_save(struct cdma *ctx, const char *phone,
    struct cdma_time *ct, const char *msg)
{
	return cmotech_ccu550_sms_save(ctx, phone, ct, msg);
}

int
cmotech_ccu650u_pb_enum(struct cdma *ctx, cdma_pb_callback cb)
{
	char *p, *s, *nxt;
	size_t sz, sz2;
	int status, error;
	char *idx, *phone, *name;

	error = send_cmd(ctx->fd, "AT$$PBRD=255\r\n");
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}

	p = ctx->buf;
	sz2 = sz;
	while (p != NULL) {
		p = get_line(p, sz2, &nxt);
		if (p == NULL)
			break;
		sz2 = sz - (p - ctx->buf);

		s = get_answer_array(p, sz2, 4, "$$PBRD");
		if (s == NULL) {
			warnx("can't read phone book");
			return (-1);
		}

		idx = s;
		s = s + strlen(s) + 1;
		phone = s;
		s = s + strlen(s) + 1;
		s = s + strlen(s) + 1;
		name = s;

		cb(idx, phone, name);

		p = nxt;
	}

	return (0);
}

/* NOTE: this modem has limitation of 20 symbols per name and phone number */
int
cmotech_ccu650u_pb_add(struct cdma *ctx, u_int index, const char *phone,
    const char *name)
{
	char *cmd;
	size_t sz;
	int status, error;

	error = snprintf(NULL, 0, "AT$$PBWR=%u,%s,2,%s\r\n", index, phone,
	    name);
	if (error == -1) {
		warnx("can't add phonebook record");
		return (-1);
	}

	cmd = (char *)malloc(error + 1);
	if (cmd == NULL) {
		warn("can't add phonebook record");
		return (-1);
	}

	error = snprintf(cmd, error + 1, "AT$$PBWR=%u,%s,2,%s\r\n", index,
	    phone, name);
	if (error == -1) {
		warnx("can't add phonebook record");
		free(cmd);
		return (-1);
	}

	error = send_cmd(ctx->fd, cmd);
	free(cmd);
	if (error != 0)
		return (-1);

	error = recv_answer(ctx->fd, &ctx->buf, &ctx->bufsz, &sz, &status);
	if (error != 0)
		return (-1);
	if (status != ANSWER_OK) {
		warnx("negative response from device");
		return (-1);
	}
	return (0);
}

int
cmotech_ccu650u_pb_delete(struct cdma *ctx, u_int index)
{
	return cmotech_ccu550_pb_delete(ctx, index);
}
