/*
 * Copyright (c) 2007 Alexey Vatchenko <av@bsdua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "util.h"

#define DEFAULT_DEVICE			"/dev/cuaU0"
#define ANSWER_BLOCK_SIZE		8096

int
open_tty(const char *dev)
{
	int fd;
	struct termios tc;
	int error;

	if (dev == NULL)
		dev = DEFAULT_DEVICE;

	fd = open(dev, O_RDWR);
	if (fd == -1) {
		warn("open");
		return (-1);
	}

	do {
		error = tcgetattr(fd, &tc);
		if (error == -1) {
			warn("tcgetattr");
			break;
		}

		tc.c_cflag &= ~(CSIZE|PARENB);
		tc.c_cflag |= CS8;
		tc.c_cflag |= CLOCAL;
		tc.c_cflag |= CRTSCTS;
		tc.c_iflag &= ~(ISTRIP|ICRNL);
		tc.c_oflag &= ~OPOST;
		tc.c_lflag &= ~(ICANON|ISIG|IEXTEN|ECHO|ECHONL);
		tc.c_cc[VMIN] = 1;
		tc.c_cc[VTIME] = 0;

		error = cfsetspeed(&tc, B115200);
		if (error == -1) {
			warn("cfsetspeed");
			break;
		}

		error = tcsetattr(fd, TCSAFLUSH, &tc);
		if (error == -1) {
			warn("cfsetspeed");
			break;
		}

		return (fd);
	} while (0);

	close(fd);
	return (-1);
}

int
send_cmd(int fd, const char *cmd)
{
	ssize_t sz;
	size_t off, len;

	len = strlen(cmd);
	off = 0;
	do {
		sz = write(fd, cmd + off, len - off);
		if (sz < 1) {
			warnx("write error");
			return (-1);
		}

		off += sz;
	} while (len != off);
	return (0);
}

/*
 * *buf must be a buffer allocated by malloc/calloc or NULL.
 * If *bufsz is not enough, *buf will be realloced.
 * The function reads device output until it reach "STATUS\r".
 */
int
recv_answer(int fd, char **buf, size_t *bufsz, size_t *rdsz, int *status)
{
	char *p;
	size_t off, soff, add;
	ssize_t readsz;

	off = 0;

	soff = 0;

	for (;;) {
		/* Reallocate buffer if it's filled */
		if (*bufsz - off == 0) {
			p = (char *)realloc(*buf, *bufsz + ANSWER_BLOCK_SIZE);
			if (p == NULL) {
				warnx("realloc");
				return (-1);
			}

			*buf = p;
			*bufsz += ANSWER_BLOCK_SIZE;
		}

		readsz = read(fd, *buf + off, *bufsz - off);
		if (readsz < 1) {
			warnx("read error");
			return (-1);
		}

		if (end_of_answer(*buf + soff, readsz + off - soff, &add,
		    status)) {
			/* XXX - it must be same as soff + add */
			*rdsz = readsz + off;
			return (0);
		}

		soff += add;
		off += readsz;
	}
}

int
end_of_answer(const char *buf, size_t bufsz, size_t *nl, int *status)
{
	size_t i, j, k;

	*nl = 0;
	for (i = 0; i < bufsz; /* nothing */) {
		j = i;
		while (buf[j] != '\r' && j < bufsz)
			j++;
		if (j == bufsz)
			return (0);
		if (j + 1 < bufsz && buf[j + 1] == '\n')
			j++;

		*nl = j + 1;

		if (*nl - i > 5 && buf[i] == 'E' && buf[i + 1] == 'R' &&
		    buf[i + 2] == 'R' && buf[i + 3] == 'O' &&
		    buf[i + 4] == 'R' && (buf[i + 5] == '\r')) {
			*status = ANSWER_ERROR;
			return (1);
		} else if (*nl - i > 2 && buf[i] == 'O' && buf[i + 1] == 'K' &&
		    (buf[i + 2] == '\r')) {
			*status = ANSWER_OK;
			return (1);
		} else if (*nl - i > 2 && buf[i] == 'A' && buf[i + 1] == 'T') {
			/* Ignore echo of AT command */
			;
		} else if (*nl - i > 1 && buf[i] == '\r') {
			/* Ignore empty lines */
			;
		} else {
			for (k = 0; k < *nl - i - 1; k++) {
				if (!isalpha(buf[i + k]) &&
				    buf[i + k] != '\r' && buf[i + k] != '\n')
					break;
			}
			if (k == *nl - i - 1) {
				*status = ANSWER_UNKNOWN;
				return (1);
			}
		}

		i = j + 1;
	}
	return (0);
}

/*
 * The function returns string that starts with $$. *n will point to the next
 * string.
 */
char *
get_line(const char *buf, size_t bufsz, char **n)
{
	char *p, *p2, *lim;

	p = (char *)buf;
	lim = p + bufsz;

	while (p < lim) {
		p2 = p;
		while (*p2 != '\r' && p2 < lim)
			p2++;
		if (p2 == lim)
			return (NULL);	/* broken string set */
		if (p2 + 1 < lim && *(p2 + 1) == '\n')
			p2++;

		if (p2 - p > 2 && *p == '$' && *(p + 1) == '$') {
			*n = p2 + 1;
			return (p);
		}

		p = p2 + 1;
	}

	return (NULL);
}

/*
 * This function is similar to get_line() but checks for $$TAG at the beginning
 * of the line. If skips lines until line is $$TAG is found.
 */
char *
get_tagged_line(const char *buf, size_t bufsz, const char *tag, char **n)
{
	char *s, *p, *nxt;
	size_t taglen, sz2;

	taglen = strlen(tag);
	sz2 = bufsz;

	s = (char *)buf;
	while ((p = get_line(s, sz2, &nxt)) != NULL) {
		if ((size_t)(nxt - p) > taglen && memcmp(p, tag, taglen) == 0) {
			*n = nxt;
			return (p);
		}
		s = nxt;
		sz2 = bufsz - (nxt - buf);
	}
	return (NULL);
}

/*
 * The function substitue (n - 1) first comma symbols and first \r or \n symbol
 * with \0 symbol. If no error occured (got enough commas) the function returns
 * pointer to the first substring.
 * NOTE that ``n'' must be greater than 0.
 */
char *
get_answer_array(const char *buf, size_t bufsz, size_t n, const char *cmd)
{
	char *p, *lim, *rv;
	size_t len;

	p = (char *)buf;
	lim = p + bufsz;
	len = strlen(cmd);

	/*
	 * Validate and skip initial "$$CMD: ".
	 */
	while (p < lim && *p != ':')
		p++;
	if (p == lim)
		return (NULL);

	if (len != (size_t)(p - buf))
		return (NULL);	/* response from other command */
	if (memcmp(cmd, buf, p - buf) != 0)
		return (NULL);	/* response from other command */

	while (++p < lim && *p == ' ')
		;		/* skip spaces */
	if (p == lim)
		return (NULL);
	rv = p;

	/*
	 * Brake buffer into substrings.
	 */
	while (n > 1) {
		while (p < lim && *p != ',')
			p++;
		if (p == lim)
			return (NULL);

		*p = '\0';
		n--;
	}
	while (p < lim && *p != '\r')
		p++;
	if (p == lim)
		return (NULL);

	*p = '\0';
	return (rv);
}

int
read_stdin(char **buf, size_t *bufsz, size_t *rdsz)
{
	char *p;
	size_t off;
	ssize_t readsz;

	off = 0;

	for (;;) {
		/* Reallocate buffer if it's filled */
		if (*bufsz - off == 0) {
			p = (char *)realloc(*buf,
			    *bufsz + ANSWER_BLOCK_SIZE + 1);
			if (p == NULL) {
				warnx("realloc");
				return (-1);
			}

			*buf = p;
			*bufsz += ANSWER_BLOCK_SIZE;
		}

		readsz = read(0, *buf + off, *bufsz - off);
		if (readsz == 0) {
			*rdsz = off;
			*(*buf + off) = '\0';
			return (0);
		} else if (readsz < 1) {
			warnx("read error");
			return (-1);
		}

		off += readsz;
	}
}
